@@ -61,19 +61,19 @@ pub fn expr_for_repr(cont: &Container) -> Result<TokenStream, syn::Error> {
 fn expr_for_field(field: &Field, allow_ref: bool) -> TokenStream {
     let (ty, type_def) = type_for_field_schema(field);
     let span = field.original.span();
-    let gen = quote!(gen);
+    let generator = quote!(generator);
 
     let mut schema_expr = if field.validation_attrs.required() {
         quote_spanned! {span=>
-            <#ty as schemars::JsonSchema>::_schemars_private_non_optional_json_schema(#gen)
+            <#ty as schemars::JsonSchema>::_schemars_private_non_optional_json_schema(#generator)
         }
     } else if allow_ref {
         quote_spanned! {span=>
-            #gen.subschema_for::<#ty>()
+            #generator.subschema_for::<#ty>()
         }
     } else {
         quote_spanned! {span=>
-            <#ty as schemars::JsonSchema>::json_schema(#gen)
+            <#ty as schemars::JsonSchema>::json_schema(#generator)
         }
     };
 
@@ -118,8 +118,8 @@ fn type_for_schema(with_attr: &WithAttr) -> (syn::Type, Option<TokenStream>) {
                         ))
                     }
 
-                    fn json_schema(gen: &mut schemars::gen::SchemaGenerator) -> schemars::schema::Schema {
-                        #fun(gen)
+                    fn json_schema(generator: &mut schemars::r#gen::SchemaGenerator) -> schemars::schema::Schema {
+                        #fun(generator)
                     }
                 }
             };
@@ -151,7 +151,7 @@ fn expr_for_external_tagged_enum<'a>(
     variants: impl Iterator<Item = &'a Variant<'a>>,
     deny_unknown_fields: bool,
 ) -> TokenStream {
-    let mut unique_names = HashSet::<String>::new();
+    let mut unique_names = HashSet::<&str>::new();
     let mut count = 0;
     let (unit_variants, complex_variants): (Vec<_>, Vec<_>) = variants
         .inspect(|v| {
@@ -178,35 +178,14 @@ fn expr_for_external_tagged_enum<'a>(
         let name = variant.name();
 
         let mut schema_expr = if variant.is_unit() && variant.attrs.with.is_none() {
-            schema_object(quote! {
-                instance_type: Some(schemars::schema::InstanceType::String.into()),
-                enum_values: Some(vec![#name.into()]),
-            })
+            quote! {
+                schemars::_private::new_unit_enum(#name)
+            }
         } else {
             let sub_schema = expr_for_untagged_enum_variant(variant, deny_unknown_fields);
-            schema_object(quote! {
-                instance_type: Some(schemars::schema::InstanceType::Object.into()),
-                object: Some(Box::new(schemars::schema::ObjectValidation {
-                    properties: {
-                        let mut props = schemars::Map::new();
-                        props.insert(#name.to_owned(), #sub_schema);
-                        props
-                    },
-                    required: {
-                        let mut required = schemars::Set::new();
-                        required.insert(#name.to_owned());
-                        required
-                    },
-                    // Externally tagged variants must prohibit additional
-                    // properties irrespective of the disposition of
-                    // `deny_unknown_fields`. If additional properties were allowed
-                    // one could easily construct an object that validated against
-                    // multiple variants since here it's the properties rather than
-                    // the values of a property that distingish between variants.
-                    additional_properties: Some(Box::new(false.into())),
-                    ..Default::default()
-                })),
-            })
+            quote! {
+                schemars::_private::new_externally_tagged_enum(#name, #sub_schema)
+            }
         };
 
         variant
@@ -227,43 +206,16 @@ fn expr_for_internal_tagged_enum<'a>(
 ) -> TokenStream {
     let mut unique_names = HashSet::new();
     let mut count = 0;
-    let set_additional_properties = if deny_unknown_fields {
-        quote! {
-            additional_properties: Some(Box::new(false.into())),
-        }
-    } else {
-        TokenStream::new()
-    };
     let variant_schemas = variants
         .map(|variant| {
             unique_names.insert(variant.name());
             count += 1;
 
             let name = variant.name();
-            let type_schema = schema_object(quote! {
-                instance_type: Some(schemars::schema::InstanceType::String.into()),
-                enum_values: Some(vec![#name.into()]),
-            });
 
-            let mut tag_schema = schema_object(quote! {
-                instance_type: Some(schemars::schema::InstanceType::Object.into()),
-                object: Some(Box::new(schemars::schema::ObjectValidation {
-                    properties: {
-                        let mut props = schemars::Map::new();
-                        props.insert(#tag_name.to_owned(), #type_schema);
-                        props
-                    },
-                    required: {
-                        let mut required = schemars::Set::new();
-                        required.insert(#tag_name.to_owned());
-                        required
-                    },
-                    // As we're creating a "wrapper" object, we can honor the
-                    // disposition of deny_unknown_fields.
-                    #set_additional_properties
-                    ..Default::default()
-                })),
-            });
+            let mut tag_schema = quote! {
+                schemars::_private::new_internally_tagged_enum(#tag_name, #name, #deny_unknown_fields)
+            };
 
             variant.attrs.as_metadata().apply_to_schema(&mut tag_schema);
 
@@ -401,9 +353,9 @@ fn variant_subschemas(unique: bool, schemas: Vec<TokenStream>) -> TokenStream {
 fn expr_for_untagged_enum_variant(variant: &Variant, deny_unknown_fields: bool) -> TokenStream {
     if let Some(with_attr) = &variant.attrs.with {
         let (ty, type_def) = type_for_schema(with_attr);
-        let gen = quote!(gen);
+        let generator = quote!(generator);
         let mut schema_expr = quote_spanned! {variant.original.span()=>
-            #gen.subschema_for::<#ty>()
+            #generator.subschema_for::<#ty>()
         };
 
         prepend_type_def(type_def, &mut schema_expr);
@@ -424,9 +376,9 @@ fn expr_for_untagged_enum_variant_for_flatten(
 ) -> Option<TokenStream> {
     if let Some(with_attr) = &variant.attrs.with {
         let (ty, type_def) = type_for_schema(with_attr);
-        let gen = quote!(gen);
+        let generator = quote!(generator);
         let mut schema_expr = quote_spanned! {variant.original.span()=>
-            <#ty as schemars::JsonSchema>::json_schema(#gen)
+            <#ty as schemars::JsonSchema>::json_schema(#generator)
         };
 
         prepend_type_def(type_def, &mut schema_expr);
@@ -443,7 +395,7 @@ fn expr_for_untagged_enum_variant_for_flatten(
 
 fn expr_for_unit_struct() -> TokenStream {
     quote! {
-        gen.subschema_for::<()>()
+        generator.subschema_for::<()>()
     }
 }
 
@@ -498,19 +450,8 @@ fn expr_for_struct(
 
             let (ty, type_def) = type_for_field_schema(field);
 
-            let maybe_insert_required = match (&default, field.validation_attrs.required()) {
-                (Some(_), _) => TokenStream::new(),
-                (None, false) => {
-                    quote! {
-                        if !<#ty as schemars::JsonSchema>::_schemars_private_is_option() {
-                            object_validation.required.insert(#name.to_owned());
-                        }
-                    }
-                }
-                (None, true) => quote! {
-                    object_validation.required.insert(#name.to_owned());
-                },
-            };
+            let has_default = default.is_some();
+            let required = field.validation_attrs.required();
 
             let metadata = SchemaMetadata {
                 read_only: field.serde_attrs.skip_deserializing(),
@@ -519,14 +460,14 @@ fn expr_for_struct(
                 ..field.attrs.as_metadata()
             };
 
-            let gen = quote!(gen);
+            let generator = quote!(generator);
             let mut schema_expr = if field.validation_attrs.required() {
                 quote_spanned! {ty.span()=>
-                    <#ty as schemars::JsonSchema>::_schemars_private_non_optional_json_schema(#gen)
+                    <#ty as schemars::JsonSchema>::_schemars_private_non_optional_json_schema(#generator)
                 }
             } else {
                 quote_spanned! {ty.span()=>
-                    #gen.subschema_for::<#ty>()
+                    #generator.subschema_for::<#ty>()
                 }
             };
 
@@ -536,8 +477,7 @@ fn expr_for_struct(
             quote! {
                 {
                     #type_def
-                    object_validation.properties.insert(#name.to_owned(), #schema_expr);
-                    #maybe_insert_required
+                    schemars::_private::insert_object_property::<#ty>(object_validation, #name, #has_default, #required, #schema_expr);
                 }
             }
         })
@@ -550,7 +490,7 @@ fn expr_for_struct(
 
             let required = field.validation_attrs.required();
 
-            let args = quote!(gen, #required);
+            let args = quote!(generator, #required);
             let mut schema_expr = quote_spanned! {ty.span()=>
                 schemars::_private::json_schema_for_flatten::<#ty>(#args)
             };
